"""Define the copiable protocol and registry."""
from typing import ClassVar, Protocol

from tentaclio.registry import URLHandlerRegistry
from tentaclio.urls import URL

from .copiers import DefaultCopier


__all__ = ["COPIER_REGISTRY"]


class Copier(Protocol):
    """Protocol of clients able to copy from the same type of client."""

    def copy(self, source: URL, dest: URL):
        """Copy from source url to dest url.

        This protocol allows clients to specialise the behaviour of copying different streams.
        """
        pass


class CopierRegistry(URLHandlerRegistry[Copier]):
    """Registry for scanners.

    The scheme expected by this registry is a compound one having
    the origin as first element and the destination as second element.
    i.e. s3+s3, file+s3, ...

    """

    def get_handler(self, scheme: str) -> Copier:
        """Get the handler for the given scheme.

        if the scheme is not set return the default copier.
        """
        if scheme not in self.registry:
            return DefaultCopier()
        return self.registry[scheme]


class _CopierRegistryHolder:
    """Module level singleton."""

    instance: ClassVar[CopierRegistry] = CopierRegistry()


COPIER_REGISTRY = _CopierRegistryHolder().instance



"""Concrete implementations of dir scanners."""
import logging
from typing import Callable, ContextManager, Iterable, Protocol

from tentaclio.urls import URL

from .scanner import DirEntry


logger = logging.getLogger(__name__)

__all__ = ["ClientDirScanner"]


class ManagedDirScanner(ContextManager, Protocol):
    """Connection based dir scanner."""

    def scandir(self, **params) -> Iterable[DirEntry]:
        """Scan based on a connection rather than a url."""
        ...


ManagedDirScannerFactory = Callable[..., ManagedDirScanner]


class ClientDirScanner:
    """DirScanner that is using a tentaclio client to scan dirs."""

    def __init__(self, client_factory: ManagedDirScannerFactory):
        """Create a new client dir scanner.

        The spected argument is class that has a context manager and a scandir method.
        i.e. tentaclio.clietns.s3_client.S3Client
        """
        self.client_factory = client_factory

    def scandir(self, url: URL) -> Iterable[DirEntry]:
        """Scan the dir-like url using the client factory to create a connection."""
        with self.client_factory(url) as client:
            return client.scandir()



"""Main entry points for fs/os like operations."""
from typing import Iterable, List, Tuple

from tentaclio import credentials

from .copier import COPIER_REGISTRY
from .remover import REMOVER_REGISTRY
from .scanner import SCANNER_REGISTRY, DirEntry


__all__ = ["scandir", "listdir", "copy", "remove", "walk"]


def scandir(url: str) -> Iterable[DirEntry]:
    """Scan a directory-like url returning its entries.

    Return an iterator of tentaclio.fs.DirEntry objects corresponding
    to the entries in the directory given by url. The entries
    are yielded in arbitrary order,
    and the special entries '.' and '..' are not included.

    More info:
    https://docs.python.org/3/library/os.html#os.scandir
    """
    authenticated = credentials.authenticate(url)
    return SCANNER_REGISTRY.get_handler(authenticated.scheme).scandir(authenticated)


def listdir(url: str) -> Iterable[str]:
    """List a directory-like url returning its entries.

    Return an iterator string containing the urls
    to the entries in the directory given by url passed as parametre. The entries
    are yielded in arbitrary order,
    and the special entries '.' and '..' are not included.

    More info:
    https://docs.python.org/3/library/os.html#os.listdir
    """
    entries = scandir(url)
    return (str(entry.url) for entry in entries)


def copy(source: str, dest: str):
    """Copy the contents of the origin URL into the dest url.

    Depending on the type of resource this can be much more performant than
    >>> with tio.open(source) as reader, tio.open(dest) as writer:
            writer.write(reader.read())
    """
    source_auth = credentials.authenticate(source)
    dest_auth = credentials.authenticate(dest)
    copier = COPIER_REGISTRY.get_handler(source_auth.scheme + "+" + dest_auth.scheme)
    copier.copy(source_auth, dest_auth)


def remove(url: str):
    """Delete the resource identified by the url."""
    authenticated = credentials.authenticate(url)
    REMOVER_REGISTRY.get_handler(authenticated.scheme).remove(authenticated)


def walk(top: str) -> Iterable[Tuple[str, str, str]]:
    """Generate the file names in a directory tree by walking the tree.

    For each directory in the tree rooted at directory top (including top itself),
    it yields a 3-tuple (dirpath, dirnames, filenames).
    """
    return _walk(top, [])


def _relativize(base: str, current: str) -> str:
    """Remove the base url from the current url."""
    if current.startswith(base):
        return current.replace(base, "", 1)
    return current


def _process_entry(
    entry: DirEntry, relative_path: str, dirs: List[str], files: List[str], to_walk: List[str]
):
    url = str(entry.url)
    if entry.is_dir:
        dirs.append(relative_path)
        to_walk.append(url)  # if a dir keep walking it.
    else:
        files.append(relative_path)
    return dirs, files, to_walk


def _walk(top: str, walked: list) -> List[Tuple[str, str, str]]:
    entries = scandir(top)
    to_walk: List[str] = []  # keep a list of dirs to walk
    dirs: List[str] = []
    files: List[str] = []

    for entry in entries:
        relative = _relativize(top, str(entry.url))
        dirs, files, to_walk = _process_entry(entry, relative, dirs, files, to_walk)

    walked.append((top, dirs, files))

    if len(to_walk):
        for new_top in to_walk:
            walked = _walk(new_top, walked)
    return walked



"""Local filesystem client."""
import os
from typing import Iterable, Union

from tentaclio import fs, protocols, urls

from . import base_client


__all__ = ["LocalFSClient"]


class LocalFSClient(base_client.BaseClient["LocalFSClient"]):
    """Local filesystem client implementation."""

    allowed_schemes = ["", "file"]

    path: str

    def __init__(self, url: Union[urls.URL, str]) -> None:
        """Create a new LocalFS client."""
        super().__init__(url)

        self.path = os.path.expanduser(self.url.path)

    def _connect(self) -> "LocalFSClient":
        return self

    def close(self) -> None:
        """Close the dummy connection to the local fs."""
        self.closed = True

    # Stream methods:

    def get(self, writer: protocols.ByteWriter, **kwargs) -> None:
        """Get the contents of the file."""
        with open(self.path, "rb") as f:
            writer.write(f.read())

    def put(self, reader: protocols.ByteReader, **kwargs) -> None:
        """Write the contents of the reader to the file."""
        with open(self.path, "wb") as f:
            f.write(bytes(reader.read()))

    # scandir related methods

    def scandir(self, **kwargs) -> Iterable[fs.DirEntry]:
        """Scan the connection url to create dir entries."""
        os_dir_entries = os.scandir(self.path)
        return map(_from_os_dir_entry, os_dir_entries)

    # remove

    def remove(self):
        """Remove the file from the local file system."""
        os.remove(self.path)


def _from_os_dir_entry(original: os.DirEntry) -> fs.DirEntry:
    return fs.DirEntry(
        url=urls.URL("file://" + os.path.abspath(original.path)),
        is_dir=bool(original.is_dir()),
        is_file=bool(original.is_file()),
    )



"""Base defines a common contract for connecting, closing, managing context for clients."""
import abc
from typing import Container, ContextManager, TypeVar, Union, cast

from tentaclio import protocols, urls


T = TypeVar("T")


class BaseClient(ContextManager[T], metaclass=abc.ABCMeta):
    """Abstract base class for clients, wrapping a connection.

    Offers context managing, connecting and closing functionalities.
    """

    url: urls.URL
    conn: protocols.Closable
    allowed_schemes: Container[str] = []
    closed: bool = True

    def __init__(self, url: Union[urls.URL, str]) -> None:
        """Create a new client based on a URL object or a string containing a url."""
        if isinstance(url, str):
            url = urls.URL(url)
        self.url = url
        if self.url.scheme not in self.allowed_schemes:
            raise ValueError(
                f"Allowed schemes for {type(self).__name__} are {self.allowed_schemes}; "
                f"found '{self.url.scheme}'"
            )

    # Context manager:

    def __enter__(self) -> T:
        """Connect to the resource when entering the context."""
        self.conn = self.connect()
        return cast(T, self)

    def __exit__(self, *args) -> None:
        """Close the connection when leaving the context."""
        self.close()

    # Connection methods:

    def connect(self) -> protocols.Closable:
        """Connect to the resource when entering the context."""
        self.closed = False
        return self._connect()

    @abc.abstractmethod
    def _connect(self) -> protocols.Closable:
        """Connect to the resource. This method needs to be overwritten by child classes."""
        ...

    def close(self) -> None:
        """Close the client connection."""
        if not self.closed:
            self.conn.close()
            self.closed = True



"""The client module defines stream and query based clients.

Stream clients allow to read and write from stream based sources with ease,
and in an interchangable manner.

Query based clients unify how to access databases leveraging from sqlalchemy.
"""
from .base_client import *  # noqa
from .ftp_client import *  # noqa
from .http_client import *  # noqa
from .importer import *  # noqa
from .local_fs_client import *  # noqa
from .sqla_client import *  # noqa



"""Base url handler registry."""
import logging
from typing import Dict, Generic, TypeVar


T = TypeVar("T")


logger = logging.getLogger(__name__)


class URLHandlerRegistry(Generic[T]):
    """Registry for url handlers based on the url scheme."""

    registry: Dict[str, T]

    def __init__(self):
        """Create a new empty registry."""
        self.registry = {}

    def register(self, scheme: str, url_handler: T):
        """Register the handler to the given scheme."""
        if scheme not in self.registry:
            logger.info(f"registering url scheme {scheme}")
            self.registry[scheme] = url_handler

    def get_handler(self, scheme: str) -> T:
        """Get the handler for the give scheme. Raise an URLError if no handler is registred."""
        if scheme not in self.registry:
            msg = f"Scheme {scheme} not found in the registry"
            logger.error(msg)
            raise KeyError(msg)
        return self.registry[scheme]

    def __contains__(self, scheme: str) -> bool:
        """Check that the scheme is in the registry."""
        return scheme in self.registry



"""Tentaclios api to deal with database clients."""
from tentaclio import credentials

from .db_registry import DB_REGISTRY, Db


__all__ = ["db"]


def db(url: str, **kwargs) -> Db:
    """Create an authenticated db client."""
    authenticated = credentials.authenticate(url)
    factory = DB_REGISTRY.get_handler(authenticated.scheme)
    return factory(authenticated, **kwargs)



"""Functionality for listing directory-like urls."""
from typing import Callable, ClassVar, ContextManager, Protocol

from tentaclio.registry import URLHandlerRegistry
from tentaclio.urls import URL


__all__ = ["REMOVER_REGISTRY", "ClientRemover"]


class Remover(ContextManager, Protocol):
    """Contract to delete the resource."""

    def remove(self):
        """Remove the underlying resource."""
        ...


class ClientRemover:
    """Wraps the client creation to get the remover."""

    def __init__(self, client_factory: Callable[..., Remover]):
        """Create the client remover."""
        self.client_factory = client_factory

    def remove(self, url: URL):
        """Build the client remover."""
        with self.client_factory(url) as client:
            client.remove()


class RemoverRegistry(URLHandlerRegistry[ClientRemover]):
    """Registry for scanners."""

    ...


class _RemoverRegistryHolder:
    """Module level singleton."""

    instance: ClassVar[RemoverRegistry] = RemoverRegistry()


REMOVER_REGISTRY = _RemoverRegistryHolder().instance



"""Public module functions."""
import functools
import logging

from tentaclio import urls

from .env import add_credentials_from_env
from .injection import CredentialsInjector
from .reader import add_credentials_from_env_file


logger = logging.getLogger(__name__)


__all__ = ["load_credentials_injector", "authenticate"]


#  Singleton
@functools.lru_cache(maxsize=1, typed=True)
def load_credentials_injector() -> CredentialsInjector:
    """Load the credentials injector fetching configuration from the environment."""
    injector = CredentialsInjector()
    injector = add_credentials_from_env(injector)
    injector = add_credentials_from_env_file(injector)
    return injector


def authenticate(url: str) -> urls.URL:
    """Authenticate url."""
    return load_credentials_injector().inject(urls.URL(url))



"""Credentials injection utilities."""
import collections
import itertools
import logging
import re
from typing import Dict, List

from tentaclio import urls


logger = logging.getLogger(__name__)

__all__ = ["CredentialsInjector"]

HOSTNAME_WILDCARD = "hostname"


class CredentialsInjector(object):
    """Allows registering and injecting credentials to urls."""

    def __init__(self):
        """Create a new credentials injector."""
        self.registry: Dict[str, List[urls.URL]] = collections.defaultdict(list)

    def register_credentials(self, url: urls.URL) -> None:
        """Register a set of credentials."""
        self.registry[url.scheme].append(url)

    def inject(self, url: urls.URL) -> urls.URL:
        """Inject credentials to the given url if they are available."""
        creds_by_scheme = self.registry[url.scheme]
        creds_by_host = _filter_by_hostname(url, creds_by_scheme)
        creds_by_path = _filter_by_path(url, creds_by_host)
        candidates = _filter_by_user(url, creds_by_path)

        if len(candidates) == 0:
            return url
        else:
            if len(candidates) > 1:
                logger.warning(
                    "The credentials for %s returned %s candidates" % (str(url), len(candidates))
                )

            # return the best candidate, i.e the one with the highest similarity to
            # the passed url
            creds = candidates[0]
            with_creds = urls.URL.from_components(
                scheme=url.scheme,
                username=creds.username,
                password=creds.password,
                hostname=creds.hostname,
                port=creds.port or url.port,
                path=url.path,
                query=url.query or creds.query,
            )
            return with_creds


_URLSimilarity = collections.namedtuple("_URLSimilarity", ["url", "similarity"])


def _filter_by_path(url: urls.URL, to_match: List[urls.URL]) -> List[urls.URL]:
    # compute the similarties between the path of the url and the list to match
    similarities = map(lambda cred: _similarity(url.path, cred.path), to_match)
    url_similarities = itertools.starmap(_URLSimilarity, zip(to_match, similarities))
    # filter out the urls without any similarity
    positives = filter(lambda url_sim: url_sim.similarity > 0, url_similarities)
    # order them by similarity the similarity the better
    positives_ordered = sorted(positives, key=lambda url_sim: url_sim.similarity, reverse=True)
    return list(map(lambda url_sim: url_sim.url, positives_ordered))


def _filter_by_hostname(url: urls.URL, to_match: List[urls.URL]) -> List[urls.URL]:
    if url.hostname != "hostname":
        return list(filter(lambda cred: cred.hostname == url.hostname, to_match))
    return to_match


def _filter_by_user(url: urls.URL, to_match: List[urls.URL]) -> List[urls.URL]:
    if url.username is None:
        return to_match
    return list(filter(lambda cred: cred.username == url.username, to_match))


PATH_DELIMITERS = "/|::"


def _similarity(path_1: str, path_2: str) -> float:
    """Compute the similarity of two strings.

    The similarity is calculated as the
    number of path elements (separated by / or ::) that they have in common.
    "/path/to/mything" and "/path/to/mything" have 3 as similarity index.
    "/path" and "/path/to/mything" have 1 as similarity index.
    "" and "" have similarity index 0.5
    "/mypath" and "" have 0.5 similarity index.

    Similarity is not symetrical:
    "" and "/mypath" have similarity 0.
    This is because
    url: "ftp://hostname/file.txt" creds: "ftp://user@hostname/" should inject to
    "ftp://user@hostname/file.txt"

    but:

    url: "db://hostname/" creds: "db://user@hostname/mydatabase" should not inject to
    "db://user@hostname/database".

    That means that the path of the url to inject to should be more specific than the credentials.
    """
    path_1 = _clean_path(path_1)
    path_2 = _clean_path(path_2)

    if not path_1 and not path_2:
        return 0.5  # non-zero similarity when path is empty

    if not path_1 or not path_2:
        return 0.5

    # split path elements

    path_1_elements = re.split(PATH_DELIMITERS, path_1)
    path_2_elements = re.split(PATH_DELIMITERS, path_2)

    return _compute_parts_similarity(path_1_elements, path_2_elements)


def _clean_path(path: str):
    path = path or ""
    return path.strip("/")


def _compute_parts_similarity(elements_1: List[str], elements_2: List[str]) -> float:
    similarity = 0
    for idx, path_element in enumerate(elements_1):
        if idx == len(elements_2):
            break
        if path_element == elements_2[idx]:
            similarity += 1
        else:
            break
    return similarity



"""Slack http hook."""
import io
import json
from typing import Optional

from tentaclio.clients import http_client


__all__ = ["SlackHook"]


class SlackHook:
    """Slack incoming web hook (POST messages)."""

    def __init__(self, url: str) -> None:
        """Create a slack hook that connects to the given url."""
        self.url = url

    def notify(
        self,
        message: str,
        channel: Optional[str] = None,
        username: Optional[str] = None,
        icon_emoji: Optional[str] = None,
    ) -> None:
        """Send a notification to slack."""
        body = self._build_request_body(
            message, channel=channel, username=username, icon_emoji=icon_emoji
        )
        stream = io.BytesIO(bytes(body, encoding="utf-8"))
        # Post streamed request
        with http_client.HTTPClient(self.url) as client:
            client.put(stream)

    # Helpers:

    @staticmethod
    def _build_request_body(
        message: str,
        channel: Optional[str] = None,
        username: Optional[str] = None,
        icon_emoji: Optional[str] = None,
    ) -> str:
        # Fetch message payload
        payload = dict(text=message)
        if channel:
            payload["channel"] = channel
        if username:
            payload["username"] = username
        if icon_emoji:
            payload["icon_emoji"] = icon_emoji
        return json.dumps(payload)



#!/usr/bin/env python
# -*- coding: utf-8 -*
import os
import pathlib
import sys

from setuptools import find_packages, setup
from setuptools.command.install import install


VERSION = "1.3.0"

REPO_ROOT = pathlib.Path(__file__).parent

with open(REPO_ROOT / "README.md", encoding="utf-8") as f:
    README = f.read()


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version."""

    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")

        if tag != VERSION:
            info = f"Git tag: {tag} does not match the version of this app: {VERSION}"
            sys.exit(info)


REQUIREMENTS = [
    # Security constraints
    "urllib3>=1.24.2",
    # Http
    "requests",
    # Sqlalchemy
    "sqlalchemy>=1.4",
    # SFTP
    "paramiko",
    # Utils
    "pandas",
    "click",
    "pyyaml",
    "importlib_metadata>3.7.0",
]

PLUGINS = {
    "s3": "tentaclio_s3",
    "athena": "tentaclio_athena",
    "postgres": "tentaclio_postgres",
    "databricks": "tentaclio_databricks>=1.0.0",
    "gdrive": "tentaclio_gdrive",
    "gs": "tentaclio_gs",
    "snowflake": "tentaclio_snowflake"
}


setup_args = dict(
    # Description
    name="tentaclio",
    version=VERSION,
    description="Unification of data connectors for distributed data tasks",
    long_description=README,
    long_description_content_type="text/markdown",
    # Credentials
    author="Octopus Energy",
    author_email="nerds@octoenergy.com",
    url="https://github.com/octoenergy/tentaclio",
    license="MIT",
    # Package data
    package_dir={"": "src"},
    packages=find_packages("src", include=["*tentaclio*"]),
    include_package_data=False,
    # Dependencies
    install_requires=REQUIREMENTS,
    extras_require=PLUGINS,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Typing :: Typed",
    ],
    cmdclass={"verify": VerifyVersionCommand},
)

if __name__ == "__main__":

    # Make install
    setup(**setup_args)



"""Read credentials from yaml file."""
import io
import logging
import os
import re

import yaml

from tentaclio import protocols, urls
from tentaclio.credentials import injection


logger = logging.getLogger(__name__)

SECRETS = "secrets"
TENTACLIO_SECRETS_FILE = "TENTACLIO__SECRETS_FILE"
ENV_VARIABLE_PATTERN = re.compile(r"\${[a-zA-Z_]+[a-zA-Z0-9_]*}")
NOT_VALID_ENV_VARIABLE_VALUES = "[${}]"


class TentaclioFileError(Exception):
    """Tentaclio secrets file errors."""

    TENTACLIO_FILE = os.getenv(TENTACLIO_SECRETS_FILE, "<unknown file>")
    ERROR_TEMPLATE = """
#########################################

Your tentaclio secrets file is malformed:

File: {tentaclio_file}

{message}

Check https://github.com/octoenergy/tentaclio#credentials-file for more info about this file
"""

    def __init__(self, message: str):
        """Intialise a new TentaclioFileError."""
        message = self.ERROR_TEMPLATE.format(
            message=message, tentaclio_file=self.TENTACLIO_FILE
        )
        super().__init__(message)


def add_credentials_from_env_file(
    injector: injection.CredentialsInjector,
) -> injection.CredentialsInjector:
    """Add credentials from a configurable yaml file.

    The path of this file is expected to be set in the environment as
    `TENTACLIO__SECRETS_FILE`
    """
    secrets_file = os.getenv(TENTACLIO_SECRETS_FILE, "")
    if secrets_file != "":
        injector = _load_from_file(injector, secrets_file)
    return injector


def _process_mark_error(error: yaml.MarkedYAMLError) -> str:
    error_str = ""
    if error.problem_mark is not None:
        error_str += str(error.problem) + "\n"
        error_str += str(error.problem_mark) + "\n"
    if error.context_mark is not None:
        error_str += str(error.context) + "\n"
        error_str += str(error.context_mark) + "\n"
    return error_str


def _load_creds_from_yaml(yaml_reader: protocols.Reader) -> dict:
    try:
        loaded_data = yaml.safe_load(io.StringIO(yaml_reader.read()))
    except yaml.MarkedYAMLError as error:
        raise TentaclioFileError(_process_mark_error(error))

    if SECRETS not in loaded_data:
        raise TentaclioFileError(
            "No secrets in yaml data. Make sure the file has a `secrets:` element"
        )
    return loaded_data[SECRETS]


def _load_from_file(
    injector: injection.CredentialsInjector, path: str
) -> injection.CredentialsInjector:
    try:
        with open(path, "r") as f:
            return add_credentials_from_reader(injector, f)
    except IOError as e:
        raise TentaclioFileError("File not found") from e


def _replace_env_variables(url: str) -> str:
    """Replace the environment variables inside the url by its value in the environment.

    Environment variables match the following format: ${ENV_VARIABLE_NAME}
    """
    env_variables = re.findall(ENV_VARIABLE_PATTERN, url)
    for env_variable in env_variables:
        env_variable_name = re.sub(NOT_VALID_ENV_VARIABLE_VALUES, "", env_variable)
        env_value = os.getenv(env_variable_name)
        if not env_value:
            raise EnvironmentError(
                f"Error while reading variable '{env_variable_name}' from the environment "
                f"when interpolating it from the secrets file. "
                f"Check that the variable exists in your environment."
            )
        url = url.replace(env_variable, env_value)
    return url


def add_credentials_from_reader(
    injector: injection.CredentialsInjector, yaml_reader: protocols.Reader
) -> injection.CredentialsInjector:
    """Read the credentials from a yml.

    The file has the follwing format:
        secrets:
            my_creds_name: http://user:password@google.com/path
            my_db: postgres://user_db:password@octoenergy.com/databasek

    """
    creds = _load_creds_from_yaml(yaml_reader)
    for name, url in creds.items():
        logger.info(f"Adding secret: {name}")
        url = _replace_env_variables(url)
        try:
            injector.register_credentials(urls.URL(url))
        except Exception as e:
            logger.error(f"Error while registering credentials {name}:{url} from file")
            raise e

    return injector
